# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Literal, NamedTuple, TypedDict


class ABIParameter(TypedDict, total=False):
    name: str
    type: str
    indexed: bool | None  # Only used in event inputs
    components: list["ABIParameter"] | None  # Used for tuple types


class ABICommon(TypedDict, total=False):
    type: str
    name: str | None
    inputs: list[ABIParameter]


class ABIFunction(ABICommon):
    outputs: list[ABIParameter]
    stateMutability: str


class ABIEvent(ABICommon):
    anonymous: bool


class ABIError(TypedDict):
    type: str
    name: str
    inputs: list[ABIParameter]


ABIElement = ABIFunction | ABIEvent | ABIError
ABI = list[ABIElement]


Environment = Literal["development", "testnet", "mainnet"]


class AgreementData(NamedTuple):
    startTime: int
    epochsNumber: int
    epochLength: int
    tokens: list[int]
    scoreFunctionIdAndProofWindowOffsetPerc: list[int]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import asyncio
from dotenv import load_dotenv
from functools import wraps
from typing import Any

import aiohttp
from dkg.constants import BLOCKCHAINS
from dkg.exceptions import (
    AccountMissing,
    NetworkNotSupported,
)
from dkg.types import URI, Address, Wei
from web3.contract import Contract
from web3.contract.contract import ContractFunction
from web3.types import TxReceipt
from web3.providers import AsyncHTTPProvider
from web3 import AsyncWeb3
from dkg.providers.blockchain.base_blockchain import BaseBlockchainProvider


class AsyncBlockchainProvider(BaseBlockchainProvider):
    def __init__(
        self,
        blockchain_id: str,
        rpc_uri: URI | None = None,
        gas_price: Wei | None = None,
        verify: bool = True,
    ):
        super().__init__(blockchain_id, rpc_uri, gas_price)

        ssl_context = None if verify else False
        self.w3 = AsyncWeb3(
            AsyncHTTPProvider(self.rpc_uri, request_kwargs={"ssl": ssl_context})
        )

        if self.blockchain_id is None:
            self.blockchain_id = f"{blockchain_id}:{self.w3.eth.chain_id}"
            if self.blockchain_id not in BLOCKCHAINS[self.environment]:
                raise NetworkNotSupported(
                    f"Network with blockchain ID {self.blockchain_id} isn't supported!"
                )

        self.gas_price_oracle = BLOCKCHAINS[self.environment][self.blockchain_id].get(
            "gas_price_oracle",
            None,
        )

        hub_address: Address = BLOCKCHAINS[self.environment][self.blockchain_id]["hub"]
        self.contracts: dict[str, Contract] = {
            "Hub": self.w3.eth.contract(
                address=hub_address,
                abi=self.abi["Hub"],
                decode_tuples=True,
            )
        }

        self.contracts_initialized = False

        load_dotenv()
        if private_key := os.environ.get("PRIVATE_KEY"):
            self.set_account(private_key)

    async def ensure_contracts_initialized(self):
        if not self.contracts_initialized:
            await self._init_contracts()
            self.contracts_initialized = True

    async def make_json_rpc_request(
        self, endpoint: str, args: dict[str, Any] = {}
    ) -> Any:
        web3_method = getattr(self.w3.eth, endpoint)

        if callable(web3_method):
            return await web3_method(**args)
        else:
            return web3_method

    @staticmethod
    def handle_updated_contract(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            contract_name = kwargs.get("contract") or (args[0] if args else None)

            try:
                return await func(self, *args, **kwargs)
            except Exception as err:
                if (
                    contract_name
                    and isinstance(contract_name, str)
                    and any(msg in str(err) for msg in ["revert", "VM Exception"])
                    and not await self._check_contract_status(contract_name)
                ):
                    is_updated = await self._update_contract_instance(contract_name)
                    if is_updated:
                        return await func(self, *args, **kwargs)
                raise err

        return wrapper

    @handle_updated_contract
    async def call_function(
        self,
        contract: str | dict[str, str],
        function: str,
        args: dict[str, Any] = {},
        state_changing: bool = False,
        gas_price: Wei | None = None,
        gas_limit: Wei | None = None,
    ) -> TxReceipt | Any:
        await self.ensure_contracts_initialized()
        if isinstance(contract, str):
            contract_name = contract
            contract_instance = self.contracts[contract_name]
        else:
            contract_name = contract["name"]
            contract_instance = self.w3.eth.contract(
                address=contract["address"],
                abi=self.abi[contract_name],
                decode_tuples=True,
            )
            self.contracts[contract_name] = contract_instance

        contract_function: ContractFunction = getattr(
            contract_instance.functions, function
        )

        if not state_changing:
            result = await contract_function(**args).call()
            if function in (
                output_named_tuples := self.output_named_tuples[contract_name]
            ):
                result = output_named_tuples[function](*result)
            return result
        else:
            if not hasattr(self, "account"):
                raise AccountMissing(
                    "State-changing transactions can be performed only with specified "
                    "account."
                )

            options = {
                "gas": gas_limit or await contract_function(**args).estimate_gas(),
            }

            gas_price = (
                self.gas_price or gas_price or await self._get_network_gas_price()
            )

            if gas_price is not None:
                options["gasPrice"] = gas_price

            tx_hash = await contract_function(**args).transact(options)
            tx_receipt = await self.w3.eth.wait_for_transaction_receipt(tx_hash)

            return tx_receipt

    async def _get_network_gas_price(self) -> Wei | None:
        if self.environment == "development":
            return None

        async def fetch_gas_price(oracle_url: str) -> Wei | None:
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(oracle_url) as response:
                        response.raise_for_status()
                        data: dict = await response.json()

                        if "result" in data:
                            return int(data["result"], 16)
                        elif "average" in data:
                            return self.w3.to_wei(data["average"], "gwei")
                        else:
                            return None
            except Exception:
                return None

        oracles = self.gas_price_oracle
        if oracles is not None:
            if isinstance(oracles, str):
                oracles = [oracles]

            for oracle_url in oracles:
                gas_price = await fetch_gas_price(oracle_url)
                if gas_price is not None:
                    return gas_price

        return None

    async def _init_contracts(self):
        init_tasks = []
        for contract in self.abi.keys():
            if contract == "Hub":
                continue
            init_tasks.append(self._update_contract_instance(contract))
        await asyncio.gather(*init_tasks)

    async def _update_contract_instance(self, contract: str) -> bool:
        [is_contract, is_storage] = await asyncio.gather(
            self.contracts["Hub"].functions.isContract(contractName=contract).call(),
            self.contracts["Hub"]
            .functions.isAssetStorage(assetStorageName=contract)
            .call(),
        )
        if is_contract or is_storage:
            self.contracts[contract] = self.w3.eth.contract(
                address=(
                    await self.contracts["Hub"]
                    .functions.getAssetStorageAddress(contract)
                    .call()
                    if contract.endswith("AssetStorage")
                    or contract.endswith("CollectionStorage")
                    else await self.contracts["Hub"]
                    .functions.getContractAddress(contract)
                    .call()
                ),
                abi=self.abi[contract],
                decode_tuples=True,
            )
            return True
        return False

    async def _check_contract_status(self, contract: str) -> bool:
        try:
            return await self.call_function(contract, "status")
        except Exception:
            return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from enum import auto, Enum
from typing import Any, Type, Dict

from dkg.dataclasses import BidSuggestionRange, HTTPRequestMethod
from dkg.exceptions import OperationFailed, OperationNotFinished
from dkg.types import AutoStrEnumUpperCase, UAL, Address, DataHexStr


@dataclass
class NodeCall:
    method: HTTPRequestMethod
    path: str
    params: dict[str, Type] = field(default_factory=dict)
    data: dict[str, Type] | Type = field(default_factory=dict)


class NodeRequest:
    info = NodeCall(method=HTTPRequestMethod.GET, path="info")
    bid_suggestion = NodeCall(
        method=HTTPRequestMethod.GET,
        path="bid-suggestion",
        params={
            "blockchain": str,
            "epochsNumber": int,
            "assertionSize": int,
            "contentAssetStorageAddress": Address,
            "firstAssertionId": DataHexStr,
            "hashFunctionId": int,
            "bidSuggestionRange": BidSuggestionRange,
        },
    )
    get_operation_result = NodeCall(
        method=HTTPRequestMethod.GET,
        path="{operation}/{operation_id}",
    )

    publish = NodeCall(
        method=HTTPRequestMethod.POST,
        path="publish",
        data={
            "datasetRoot": str,
            "dataset": dict[str, list[str]],
            "blockchain": str,
            "hashFunctionId": int,
            "minimumNumberOfNodeReplications": int,
        },
    )

    ask = NodeCall(
        method=HTTPRequestMethod.POST,
        path="ask",
        params={
            "ual": UAL,
            "minimumNumberOfNodeReplications": int,
        },
    )

    finality_status = NodeCall(
        method=HTTPRequestMethod.GET,
        path="finality",
        params={"ual": UAL},
    )

    get = NodeCall(
        method=HTTPRequestMethod.POST,
        path="get",
        data={
            "id": UAL,
            "contentType": str,
            "includeMetadata": bool,
            "hashFunctionId": int,
            "paranetUAL": UAL,
            "subjectUAL": UAL,
        },
    )

    query = NodeCall(
        method=HTTPRequestMethod.POST,
        path="direct-query",
        data={
            "query": str,
            "type": str,
            "repository": str | None,
            "paranetUAL": str | None,
        },
    )


class LocalStoreOperationStatus(AutoStrEnumUpperCase):
    LOCAL_STORE_INIT_START = auto()
    LOCAL_STORE_INIT_END = auto()
    LOCAL_STORE_START = auto()
    LOCAL_STORE_END = auto()


class PublishOperationStatus(Enum):
    VALIDATING_PUBLISH_ASSERTION_REMOTE_START = auto()
    VALIDATING_PUBLISH_ASSERTION_REMOTE_END = auto()
    INSERTING_ASSERTION = auto()
    PUBLISHING_ASSERTION = auto()
    PUBLISH_START = auto()
    PUBLISH_INIT_START = auto()
    PUBLISH_INIT_END = auto()
    PUBLISH_LOCAL_STORE_START = auto()
    PUBLISH_LOCAL_STORE_END = auto()
    PUBLISH_REPLICATE_START = auto()
    PUBLISH_REPLICATE_END = auto()
    PUBLISH_END = auto()


class StoreTypes(AutoStrEnumUpperCase):
    TRIPLE = auto()
    TRIPLE_PARANET = auto()
    PENDING = auto()


class GetOperationStatus(AutoStrEnumUpperCase):
    ASSERTION_EXISTS_LOCAL_START = auto()
    ASSERTION_EXISTS_LOCAL_END = auto()
    GET_START = auto()
    GET_INIT_START = auto()
    GET_INIT_END = auto()
    GET_LOCAL_START = auto()
    GET_LOCAL_END = auto()
    GET_REMOTE_START = auto()
    GET_REMOTE_END = auto()
    GET_FETCH_FROM_NODES_START = auto()
    GET_FETCH_FROM_NODES_END = auto()
    GET_END = auto()


class QueryOperationStatus(AutoStrEnumUpperCase):
    QUERY_INIT_START = auto()
    QUERY_INIT_END = auto()
    QUERY_START = auto()
    QUERY_END = auto()


class OperationStatus(AutoStrEnumUpperCase):
    PENDING = auto()
    FAILED = auto()
    COMPLETED = auto()
    FIND_NODES_START = auto()
    FIND_NODES_END = auto()
    FIND_NODES_LOCAL_START = auto()
    FIND_NODES_LOCAL_END = auto()
    FIND_NODES_OPEN_CONNECTION_START = auto()
    FIND_NODES_OPEN_CONNECTION_END = auto()
    FIND_NODES_CREATE_STREAM_START = auto()
    FIND_NODES_CREATE_STREAM_END = auto()
    FIND_NODES_SEND_MESSAGE_START = auto()
    FIND_NODES_SEND_MESSAGE_END = auto()
    DIAL_PROTOCOL_START = auto()
    DIAL_PROTOCOL_END = auto()
    LOCAL_STORE = LocalStoreOperationStatus
    PUBLISH = PublishOperationStatus
    GET = GetOperationStatus
    QUERY = QueryOperationStatus


def validate_operation_status(operation_result: dict[str, Any]) -> None:
    try:
        status = OperationStatus(operation_result["status"])
    except ValueError:
        raise OperationNotFinished("Operation isn't finished")

    match status:
        case OperationStatus.COMPLETED:
            return
        case OperationStatus.FAILED:
            raise OperationFailed(
                f"Operation failed! {operation_result['data']['errorType']}: "
                f"{operation_result['data']['errorMessage']}."
            )
        case _:
            raise OperationNotFinished("Operation isn't finished")


def get_operation_status_object(
    operation_result: Dict[str, Any], operation_id: str
) -> Dict[str, Any]:
    """
    Creates an operation status object from operation result and ID.

    Args:
        operation_result: Dictionary containing operation result data
        operation_id: The ID of the operation

    Returns:
        Dictionary containing operation status information
    """
    # Check if error_type exists in operation_result.data
    operation_data = (
        {"status": operation_result.get("status"), **operation_result.get("data")}
        if operation_result.get("data", {}).get("errorType")
        else {"status": operation_result.get("status")}
    )

    return {"operationId": operation_id, **operation_data}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import wraps


from dkg.modules.asset.async_asset import AsyncKnowledgeAsset
from dkg.modules.graph.async_graph import AsyncGraph
from dkg.managers.async_manager import AsyncRequestManager
from dkg.modules.async_module import AsyncModule
from dkg.modules.node.async_node import AsyncNode
from dkg.types import UAL, Address, ChecksumAddress
from dkg.utils.ual import format_ual, parse_ual
from dkg.services.input_service import InputService
from dkg.providers.blockchain.async_blockchain import AsyncBlockchainProvider
from dkg.providers.node.async_node_http import AsyncNodeHTTPProvider
from dkg.services.node_services.async_node_service import AsyncNodeService
from dkg.services.blockchain_services.async_blockchain_service import (
    AsyncBlockchainService,
)


class AsyncDKG(AsyncModule):
    asset: AsyncKnowledgeAsset
    node: AsyncNode
    graph: AsyncGraph

    def __init__(
        self,
        node_provider: AsyncNodeHTTPProvider,
        blockchain_provider: AsyncBlockchainProvider,
        config: dict = {},
    ):
        self.manager = AsyncRequestManager(node_provider, blockchain_provider)

        self.initialize_services(config)

        modules = {
            "asset": AsyncKnowledgeAsset(
                self.manager,
                self.input_service,
                self.node_service,
                self.blockchain_service,
            ),
            "node": AsyncNode(self.manager, self.node_service),
            "graph": AsyncGraph(self.manager, self.input_service, self.node_service),
        }
        self._attach_modules(modules)

        self._setup_backwards_compatibility()

    def _setup_backwards_compatibility(self):
        # Create async wrapper methods
        async def graph_get(*args, **kwargs):
            return await self.asset.get(*args, **kwargs)

        async def graph_create(*args, **kwargs):
            return await self.asset.create(*args, **kwargs)

        # Attach methods to graph
        self.graph.get = graph_get
        self.graph.create = graph_create

    def initialize_services(self, config: dict = {}):
        self.input_service = InputService(self.manager, config)
        self.node_service = AsyncNodeService(self.manager)
        self.blockchain_service = AsyncBlockchainService(self.manager)

    @staticmethod
    @wraps(format_ual)
    def format_ual(
        blockchain: str, contract_address: Address | ChecksumAddress, token_id: int
    ) -> UAL:
        return format_ual(blockchain, contract_address, token_id)

    @staticmethod
    @wraps(parse_ual)
    def parse_ual(ual: UAL) -> dict[str, str | Address | int]:
        return parse_ual(ual)

    @property
    def node_provider(self) -> AsyncNodeHTTPProvider:
        return self.manager.node_provider

    @node_provider.setter
    def node_provider(self, node_provider: AsyncNodeHTTPProvider) -> None:
        self.manager.node_provider = node_provider

    @property
    def blockchain_provider(self) -> AsyncBlockchainProvider:
        return self.manager.blockchain_provider

    @blockchain_provider.setter
    def blockchain_provider(self, blockchain_provider: AsyncBlockchainProvider) -> None:
        self.manager.blockchain_provider = blockchain_provider



from abc import ABC, abstractmethod
from dkg.dataclasses import HTTPRequestMethod, NodeResponseDict
from typing import Any
from dkg.types import URI


class BaseNodeHTTPProvider(ABC):
    def __init__(
        self,
        endpoint_uri: URI | str,
        api_version: str = "v1",
        auth_token: str | None = None,
    ):
        self.url = f"{URI(endpoint_uri)}/{api_version}"
        self.headers = {"Authorization": f"Bearer {auth_token}"} if auth_token else {}

    @abstractmethod
    async def make_request(
        self,
        method: HTTPRequestMethod,
        path: str,
        params: dict[str, Any] = {},
        data: dict[str, Any] = {},
    ) -> NodeResponseDict:
        raise NotImplementedError("Subclasses must implement make_request")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import hashlib
import json

from dkg.types import Address, NQuads
from eth_abi.packed import encode_packed


def generate_assertion_metadata(assertion: NQuads) -> dict[str, int]:
    return {
        "size": len(json.dumps(assertion, separators=(",", ":")).encode("utf-8")),
        "triples_number": len(assertion),
        "chunks_number": len(assertion),  # TODO: Change when chunking introduced
    }


def generate_keyword(contract_address: Address, assertion_id: bytes) -> bytes:
    return encode_packed(
        ["address", "bytes32"],
        [contract_address, assertion_id],
    )


def generate_agreement_id(
    contract_address: Address,
    token_id: int,
    keyword: bytes,
) -> bytes:
    return hashlib.sha256(
        encode_packed(
            ["address", "uint256", "bytes"],
            [contract_address, token_id, keyword],
        )
    ).digest()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time
from functools import wraps
from typing import Any, Callable

from dkg.exceptions import NodeRequestError


def retry(
    catch: Exception, max_retries: int, base_delay: int, backoff: float
) -> Callable[[Callable], Callable]:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            _delay = base_delay

            for _ in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except catch:
                    time.sleep(_delay)
                    _delay *= backoff

            raise NodeRequestError(
                f"Failed executing {func.__name__} after {max_retries} retries."
            )

        return wrapper

    return decorator



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Type, Dict, Any

from dkg.dataclasses import ParanetIncentivizationType
from dkg.types import Address, HexStr, Wei


@dataclass
class JSONRPCRequest:
    endpoint: str
    args: dict[str, Type] = field(default_factory=dict)


@dataclass
class ContractInteraction:
    contract: str | None = None
    function: str = field(default_factory=str)
    args: dict[str, Type] = field(default_factory=dict)

    def __post_init__(self):
        if not self.function:
            raise ValueError(
                "'function' is a required field and cannot be None or empty"
            )


@dataclass
class ContractTransaction(ContractInteraction):
    gas_price: Wei | None = None
    gas_limit: Wei | None = None


@dataclass
class ContractCall(ContractInteraction):
    pass


@dataclass
class KnowledgeCollectionResult:
    knowledge_collection_id: int
    receipt: Dict[str, Any]


@dataclass
class AllowanceResult:
    allowance_increased: bool
    allowance_gap: int


class BlockchainRequest:
    chain_id = JSONRPCRequest("chain_id")
    get_block = JSONRPCRequest("get_block", args={"block_identifier": str | int})

    get_contract_address = ContractCall(
        contract="Hub",
        function="getContractAddress",
        args={"contractName": str},
    )
    get_asset_storage_address = ContractCall(
        contract="Hub",
        function="getAssetStorageAddress",
        args={"assetStorageName": str},
    )

    key_is_operational_wallet = ContractCall(
        contract="IdentityStorage",
        function="keyHasPurpose",
        args={"identityId": int, "_key": Address, "_purpose": int},
    )

    time_until_next_epoch = ContractCall(
        contract="Chronos",
        function="timeUntilNextEpoch",
        args={},
    )

    epoch_length = ContractCall(
        contract="Chronos",
        function="epochLength",
        args={},
    )

    get_stake_weighted_average_ask = ContractCall(
        contract="AskStorage",
        function="getStakeWeightedAverageAsk",
        args={},
    )

    allowance = ContractCall(
        contract="Token",
        function="allowance",
        args={"owner": Address, "spender": Address},
    )
    increase_allowance = ContractTransaction(
        contract="Token",
        function="increaseAllowance",
        args={"spender": Address, "addedValue": Wei},
    )
    decrease_allowance = ContractTransaction(
        contract="Token",
        function="decreaseAllowance",
        args={"spender": Address, "subtractedValue": Wei},
    )

    burn_asset = ContractTransaction(
        contract="ContentAsset",
        function="burnAsset",
        args={"tokenId": int},
    )
    extend_asset_storing_period = ContractTransaction(
        contract="ContentAsset",
        function="extendAssetStoringPeriod",
        args={"tokenId": int, "epochsNumber": int, "tokenAmount": int},
    )

    transfer_asset = ContractTransaction(
        contract="ContentAssetStorage",
        function="transferFrom",
        args={"from": Address, "to": Address, "tokenId": int},
    )
    get_latest_assertion_id = ContractCall(
        contract="ContentAssetStorage",
        function="getLatestAssertionId",
        args={"tokenId": int},
    )
    owner_of = ContractCall(
        contract="ContentAssetStorage",
        function="ownerOf",
        args={"tokenId": int},
    )

    get_assertion_size = ContractCall(
        contract="AssertionStorage",
        function="getAssertionSize",
        args={"assertionId": bytes | HexStr},
    )

    # Identity
    get_identity_id = ContractCall(
        contract="IdentityStorage",
        function="getIdentityId",
        args={"operational": Address},
    )

    # Paranets
    register_paranet = ContractTransaction(
        contract="Paranet",
        function="registerParanet",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "paranetName": str,
            "paranetDescription": str,
            "nodesAccessPolicy": int,
            "minersAccessPolicy": int,
        },
    )

    add_paranet_curated_nodes = ContractTransaction(
        contract="Paranet",
        function="addParanetCuratedNodes",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "identityIds": list[int],
        },
    )

    remove_paranet_curated_nodes = ContractTransaction(
        contract="Paranet",
        function="removeParanetCuratedNodes",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "identityIds": list[int],
        },
    )

    request_paranet_curated_node_access = ContractTransaction(
        contract="Paranet",
        function="requestParanetCuratedNodeAccess",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
        },
    )

    approve_curated_node = ContractTransaction(
        contract="Paranet",
        function="approveCuratedNode",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "identityId": int,
        },
    )

    reject_curated_node = ContractTransaction(
        contract="Paranet",
        function="rejectCuratedNode",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "identityId": int,
        },
    )

    get_curated_nodes = ContractCall(
        contract="ParanetsRegistry",
        function="getCuratedNodes",
        args={"paranetId": HexStr},
    )

    add_paranet_curated_miners = ContractTransaction(
        contract="Paranet",
        function="addParanetCuratedMiners",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "minerAddresses": list[Address],
        },
    )

    remove_paranet_curated_miners = ContractTransaction(
        contract="Paranet",
        function="removeParanetCuratedMiners",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "minerAddresses": list[Address],
        },
    )

    request_paranet_curated_miner_access = ContractTransaction(
        contract="Paranet",
        function="requestParanetCuratedMinerAccess",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
        },
    )

    approve_curated_miner = ContractTransaction(
        contract="Paranet",
        function="approveCuratedMiner",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "minerAddress": Address,
        },
    )

    reject_curated_miner = ContractTransaction(
        contract="Paranet",
        function="rejectCuratedMiner",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "minerAddress": Address,
        },
    )

    get_knowledge_miners = ContractCall(
        contract="ParanetsRegistry",
        function="getKnowledgeMiners",
        args={"paranetId": HexStr},
    )

    add_paranet_services = ContractTransaction(
        contract="Paranet",
        function="addParanetServices",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "services": dict[str, Address | int],
        },
    )
    register_paranet_service = ContractTransaction(
        contract="Paranet",
        function="registerParanetService",
        args={
            "paranetServiceKAStorageContract": Address,
            "paranetServiceKATokenId": int,
            "paranetServiceName": str,
            "paranetServiceDescription": str,
            "paranetServiceAddresses": list[Address],
        },
    )
    submit_knowledge_collection = ContractTransaction(
        contract="Paranet",
        function="submitKnowledgeCollection",
        args={
            "paranetKCStorageContract": Address,
            "paranetKnowledgeCollectionId": int,
            "knowledgeCollectionStorageContract": Address,
            "knowledgeCollectionTokenId": int,
        },
    )

    deploy_neuro_incentives_pool = ContractTransaction(
        contract="ParanetIncentivesPoolFactory",
        function="deployNeuroIncentivesPool",
        args={
            "paranetKAStorageContract": Address,
            "paranetKATokenId": int,
            "tracToNeuroEmissionMultiplier": float,
            "paranetOperatorRewardPercentage": float,
            "paranetIncentivizationProposalVotersRewardPercentage": float,
        },
    )
    get_incentives_pool_address = ContractCall(
        contract="ParanetsRegistry",
        function="getIncentivesPoolAddress",
        args={
            "paranetId": HexStr,
            "incentivesPoolType": ParanetIncentivizationType,
        },
    )

    get_updating_knowledge_collection_states = ContractCall(
        contract="ParanetKnowledgeMinersRegistry",
        function="getUpdatingKnowledgeCollectionStates",
        args={
            "miner": Address,
            "paranetId": HexStr,
        },
    )

    is_knowledge_miner_registered = ContractCall(
        contract="ParanetsRegistry",
        function="isKnowledgeMinerRegistered",
        args={
            "paranetId": HexStr,
            "knowledgeMinerAddress": Address,
        },
    )
    is_proposal_voter = ContractCall(
        function="isProposalVoter",
        args={"addr": Address},
    )

    get_claimable_knowledge_miner_reward_amount = ContractCall(
        function="getClaimableKnowledgeMinerRewardAmount",
        args={},
    )
    get_claimable_all_knowledge_miners_reward_amount = ContractCall(
        function="getClaimableAllKnowledgeMinersRewardAmount",
        args={},
    )
    claim_knowledge_miner_reward = ContractTransaction(
        function="claimKnowledgeMinerReward",
        args={},
    )

    get_claimable_paranet_operator_reward_amount = ContractCall(
        function="getClaimableParanetOperatorRewardAmount",
        args={},
    )
    claim_paranet_operator_reward = ContractTransaction(
        function="claimParanetOperatorReward",
        args={},
    )

    get_claimable_proposal_voter_reward_amount = ContractCall(
        function="getClaimableProposalVoterRewardAmount",
        args={},
    )
    get_claimable_all_proposal_voters_reward_amount = ContractCall(
        function="getClaimableAllProposalVotersRewardAmount",
        args={},
    )
    claim_incentivization_proposal_voter_reward = ContractTransaction(
        function="claimIncentivizationProposalVoterReward",
        args={},
    )

    create_knowledge_collection = ContractTransaction(
        contract="KnowledgeCollection",
        function="createKnowledgeCollection",
        args={
            "publishOperationId": str,
            "merkleRoot": bytes,
            "knowledgeAssetsAmount": int,
            "byteSize": int,
            "epochs": int,
            "tokenAmount": int,
            "isImmutable": bool,
            "paymaster": Address,
            "publisherNodeIdentityId": int,
            "publisherNodeR": bytes,
            "publisherNodeVS": bytes,
            "identityIds": list[int],
            "r": list[bytes],
            "vs": list[bytes],
        },
    )

    mint_knowledge_collection = ContractTransaction(
        contract="Paranet",
        function="mintKnowledgeCollection",
        args={
            "paranetKCStorageContract": Address,
            "paranetKCTokenId": int,
            "knowledgeAssetArgs": dict,
        },
    )
